/*
 * Copyright (c) 2016-2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Mar  2 2016
 */
#include <sys/time.h>
#include <math.h>

#pragma GCC diagnostic ignored "-Wunknown-pragmas"

#define S_FUNCTION_NAME genomix_block
#define S_FUNCTION_LEVEL 2

#include "simstruc.h"

#include "libgenomix.h"


/* --- local data ---------------------------------------------------------- */

enum {
  GENOMIX_PORT, GENOMIX_RQST
};

enum {
  GENOMIX_PARAM_KIND = 0,
  GENOMIX_PARAM_HOST,
  GENOMIX_PARAM_PATH,
  GENOMIX_PARAM_COMPONENT,
  GENOMIX_PARAM_INSTANCE,
  GENOMIX_PARAM_PORT,
  GENOMIX_PARAM_SUBPORT,
  GENOMIX_PARAM_TIMETYPE,
  GENOMIX_PARAM_PERIOD,
  GENOMIX_PARAM_SIGLIST,
  GENOMIX_NPARAM
};

enum {
  GENOMIX_PERIODIC, GENOMIX_TRIG_LEVEL, GENOMIX_TRIG_RISING
};

struct genomix_siglist {
  enum { GENOMIX_END, GENOMIX_VALUE, GENOMIX_SIZE } type;
  char *name;
  int size;
};

struct genomix_dict {
  char *key;
  double value;
  struct genomix_dict *next;
};

struct genomix_port {
  int id;
  int kind, trigtype, trignum;
  bool trigger;
  char component[64], instance[64];
  char url[1024];
  mxArray *meta;
  char *json;
  size_t jsonlen;
  struct genomix_siglist *siglist;
  struct genomix_dict *dict;		/* holding enums */
};

static struct genomix_siglist *	getsiglist(SimStruct *S);
static int			s2p_value(const mxArray *data,
                                          int type, const char *field,
                                          struct genomix_dict *dict,
                                          double **out, int *remain,
                                          char *emsg, size_t emsgsize);
static int			p2s_value(mxArray *data,
                                          int type, const char *field,
                                          struct genomix_dict *dict,
                                          const double **in, int *remain,
                                          char *emsg, size_t emsgsize);

static mxArray *		m2s_input(const mxArray *input,
                                          struct genomix_dict **dict);
static mxArray *		m2s_value(const mxArray *format,
                                          struct genomix_dict **dict);

/*  --- mdlInitializeSizes ------------------------------------------------- */

/* Abstract:
 *    Determine the S-function block's characteristics:
 *    number of inputs, outputs, states, etc.
 */
static void
mdlInitializeSizes(SimStruct *S)
{
  struct genomix_siglist *siglist;
  const mxArray *p;
  size_t i, ninports, noutports;
  char kind[16], ttype[16];

  /* parameters */
  ssSetNumSFcnParams(S, GENOMIX_NPARAM);
  if (ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S))
    return; /* error reported by the Simulink engine */

  p = ssGetSFcnParam(S, GENOMIX_PARAM_KIND);
  if (!p) { ssSetErrorStatus(S, "invalid arguments"); return; }
  mxGetString(p, kind, sizeof(kind));

  p = ssGetSFcnParam(S, GENOMIX_PARAM_TIMETYPE);
  if (!p) { ssSetErrorStatus(S, "invalid arguments"); return; }
  mxGetString(p, ttype, sizeof(ttype));

  /* ports */
  siglist = getsiglist(S);
  if (siglist)
    for(i = 0; siglist[i].type != GENOMIX_END; i++)
      /* empty body */;
  else
    i = 0;

  ninports = noutports = 0;
  if (!strcmp(kind, "port")) noutports = i; else ninports = i;

  if (strcmp(ttype, "periodic")) {
    if (!ssSetNumInputPorts(S, 1 + ninports)) return;
    ssSetInputPortWidth(S, ninports, 1);
    ssSetInputPortDataType(S, ninports, SS_BOOLEAN);
    ssSetInputPortDirectFeedThrough(S, ninports, 1);
  } else {
    if (!ssSetNumInputPorts(S, ninports)) return;
  }
  if (!ssSetNumOutputPorts(S, noutports)) return;

  for(i = 0; i < ninports; i++) {
    ssSetInputPortWidth(S, i, siglist[i].size);
    ssSetInputPortDataType(S, i, SS_DOUBLE);
    ssSetInputPortDirectFeedThrough(S, i, 1);
    ssSetInputPortRequiredContiguous(S, i, 1);
  }
  for(i = 0; i < noutports; i++) {
    ssSetOutputPortWidth(S, i, siglist[i].size);
    ssSetOutputPortDataType(S, i, SS_DOUBLE);
  }

  for(i = 0; siglist[i].type != GENOMIX_END; i++)
    free(siglist[i].name);
  if (siglist) free(siglist);

  /* work vectors */
  ssSetNumPWork(S, 1);

  /* timings */
  ssSetNumSampleTimes(S, 1);
}

static void
mdlInitializeSampleTimes(SimStruct *S)
{
  const mxArray *p;
  char ttype[64];

  p = ssGetSFcnParam(S, GENOMIX_PARAM_TIMETYPE);
  if (!p) { ssSetErrorStatus(S, "invalid arguments"); return; }
  mxGetString(p, ttype, sizeof(ttype));

  if (strcmp(ttype, "periodic")) {
    ssSetSampleTime(S, 0, INHERITED_SAMPLE_TIME);
  } else {
    p = ssGetSFcnParam(S, GENOMIX_PARAM_PERIOD);
    if (!p) { ssSetErrorStatus(S, "invalid arguments"); return; }

    ssSetSampleTime(S, 0, mxGetScalar(p));
  }
  ssSetOffsetTime(S, 0, 0.0);
}

#define MDL_START
static void
mdlStart(SimStruct *S)
{
  struct genomix_port *genomix;
  char key[64], emsg[256];
  const mxArray *p;
  mxArray *meta;
  const char *r;
  char *s;
  int i;

  ssSetPWorkValue(S, 0, NULL);
  genomix = malloc(sizeof(*genomix));
  if (!genomix) {
    ssSetErrorStatus(S, "not enough memory");
    return;
  }
  genomix->id = -1;
  genomix->meta = NULL;
  genomix->json = malloc(genomix->jsonlen = 128);
  genomix->dict = NULL;

  /* signal list */
  genomix->siglist = getsiglist(S);
  if (!genomix->siglist) goto error;

  /* parameters */
  p = ssGetSFcnParam(S, GENOMIX_PARAM_KIND);
  if (!p) { ssSetErrorStatus(S, "invalid arguments"); goto error; }
  mxGetString(p, key, sizeof(key));
  if (strcmp(key, "port"))
    genomix->kind = GENOMIX_RQST;
  else
    genomix->kind = GENOMIX_PORT;

  p = ssGetSFcnParam(S, GENOMIX_PARAM_TIMETYPE);
  if (!p) { ssSetErrorStatus(S, "invalid arguments"); goto error; }
  mxGetString(p, key, sizeof(key));
  if (strcmp(key, "periodic")) {
    if (genomix->kind == GENOMIX_PORT)
      i = 0;
    else for(i = 0; genomix->siglist[i].type != GENOMIX_END; i++)
           /* empty body */;
    genomix->trignum = i;
    if (strstr(key, "rising"))
      genomix->trigtype = GENOMIX_TRIG_RISING;
    else
      genomix->trigtype = GENOMIX_TRIG_LEVEL;
  } else {
    genomix->trignum = -1;
    genomix->trigtype = GENOMIX_PERIODIC;
  }
  genomix->trigger = false;

  /* connect */
  p = ssGetSFcnParam(S, GENOMIX_PARAM_HOST);
  if (!p) { ssSetErrorStatus(S, "invalid arguments"); goto error; }
  s = mxArrayToString(p);
  genomix->id = genomix_connect(s, "", emsg, sizeof(emsg));
  mxFree(s);
  if (genomix->id < 0) { ssSetErrorStatus(S, strdup(emsg)); goto error; }

  /* load */
  p = ssGetSFcnParam(S, GENOMIX_PARAM_COMPONENT);
  if (!p) { ssSetErrorStatus(S, "invalid arguments"); goto error; }
  mxGetString(p, genomix->component, sizeof(genomix->component));

  if (*genomix->component) {
    const char *args[] = { "argv", NULL, NULL };
    char instance[128];
    char path[1024];

    p = ssGetSFcnParam(S, GENOMIX_PARAM_PATH);
    s = p ? mxArrayToString(p) : NULL;
    if (s && *s) {
      snprintf(path, sizeof(path), "/load/%s/%s", s, genomix->component);
      mxFree(s);
    } else
      snprintf(path, sizeof(path), "/load/%s", genomix->component);

    p = ssGetSFcnParam(S, GENOMIX_PARAM_INSTANCE);
    if (!p) { ssSetErrorStatus(S, "invalid arguments"); goto error; }
    mxGetString(p, genomix->instance, sizeof(genomix->instance));
    if (*genomix->instance) {
      snprintf(instance, sizeof(instance), "--name %s", genomix->instance);
      args[1] = instance;
    } else
      snprintf(genomix->instance, sizeof(genomix->instance),
               "%s", genomix->component);

    r = genomix_get(genomix->id, path, args, emsg, sizeof(emsg));
    if (!r) { ssSetErrorStatus(S, strdup(emsg)); goto error; }

    /* info */
    snprintf(path, sizeof(path), "/module/%s/info", genomix->instance);
    r = genomix_get(genomix->id, path, args, emsg, sizeof(emsg));
    if (!r) { ssSetErrorStatus(S, strdup(emsg)); goto error; }

    meta = j2s_value(&r, emsg, sizeof(emsg));
    if (!meta) { ssSetErrorStatus(S, strdup(emsg)); goto error; }

  } else {
    ssSetErrorStatus(S, "invalid arguments"); goto error;
  }

  /* port/service url */
  p = ssGetSFcnParam(S, GENOMIX_PARAM_PORT);
  if (!p) { ssSetErrorStatus(S, "invalid arguments"); goto error; }
  s = mxArrayToString(p);
  if (s) {
    if (genomix->kind == GENOMIX_PORT) {
      p = ssGetSFcnParam(S, GENOMIX_PARAM_SUBPORT);
      if (!p) { ssSetErrorStatus(S, "invalid arguments"); goto error; }
      mxGetString(p, key, sizeof(key));
      if (*key)
        snprintf(genomix->url, sizeof(genomix->url), "/module/%s/read/%s/%s",
                 genomix->instance, s, key);
      else
        snprintf(genomix->url, sizeof(genomix->url), "/module/%s/read/%s",
                 genomix->instance, s);
    } else
      snprintf(genomix->url, sizeof(genomix->url), "/module/%s/send/%s",
               genomix->instance, s);

    p = mxGetField(
      meta, 0, genomix->kind == GENOMIX_PORT ? "ports" : "services");
    if (!p) { ssSetErrorStatus(S, "invalid metadata"); goto error; }

    p = mxGetField(p, 0, s);
    if (!p) {
      snprintf(emsg, sizeof(emsg),
               "invalid object `%s' in `%s'", s, genomix->instance);
      ssSetErrorStatus(S, strdup(emsg));
      goto error;
    }

    p = mxGetField(
      p, 0, genomix->kind == GENOMIX_PORT ? "data" : "input");
    if (!p) { ssSetErrorStatus(S, "invalid metadata"); goto error; }

    genomix->meta = m2s_input(p, &genomix->dict);
    mexMakeArrayPersistent(genomix->meta);

    mxDestroyArray(meta);
    mxFree(s);
  } else {
    ssSetErrorStatus(S, "invalid arguments"); goto error;
  }

  /* done */
  ssSetPWorkValue(S, 0, genomix);
  return;

error:
  if (genomix->id >=0)
    genomix_disconnect(genomix->id);
  if (genomix->meta)
    mxDestroyArray(genomix->meta);
  if (genomix->json)
    free(genomix->json);
  if (genomix->siglist) {
    for(i = 0; genomix->siglist[i].type != GENOMIX_END; i++)
      free(genomix->siglist[i].name);
    free(genomix->siglist);
  }
  if (genomix) free(genomix);
}

static void
mdlOutputs(SimStruct *S, int_T tid)
{
  struct genomix_port *genomix = ssGetPWorkValue(S, 0);
  struct genomix_siglist *siglist;
  UNUSED_ARG(tid); /* not used in single tasking mode */
  char emsg[256];
  const char *r;
  int i, size;

  /* trigger */
  if (genomix->trigtype != GENOMIX_PERIODIC) {
    InputPtrsType t = ssGetInputPortSignalPtrs(S, genomix->trignum);
    bool v = *((InputBooleanPtrsType)t)[0];
    bool r = true;

    switch (genomix->trigtype) {
      case GENOMIX_TRIG_LEVEL: if (!v) r = false; break;
      case GENOMIX_TRIG_RISING: if (!v || genomix->trigger) r = false; break;
    }
    genomix->trigger = v;
    if (!r) return;
  }

  /* get inputs */
  if (genomix->kind == GENOMIX_RQST) {
    char *ejson = genomix->json;

    /* import from ports */
    for(siglist = genomix->siglist, i = 0;
        siglist->type != GENOMIX_END;
        siglist++, i++) {
      const double *port = ssGetInputPortRealSignal(S, i);
      size = siglist->size;
      if (p2s_value(genomix->meta, siglist->type, siglist->name, genomix->dict,
                    &port, &size, emsg, sizeof(emsg)))
        ssSetErrorStatus(S, strdup(emsg));
    }

    /* convert struct to json */
    s2j_value(genomix->meta, &genomix->json, &ejson, &genomix->jsonlen);
  }

  /* get url */
  if (genomix->kind == GENOMIX_RQST) {
    const char *args[] = { "oneway", NULL, "input", genomix->json, NULL };

    r = genomix_get(genomix->id, genomix->url, args, emsg, sizeof(emsg));
  } else
    r = genomix_get(genomix->id, genomix->url, NULL, emsg, sizeof(emsg));
  if (!r) { ssSetErrorStatus(S, strdup(emsg)); return; }

  /* assign outputs */
  if (genomix->kind == GENOMIX_PORT) {
    /* convert json to struct */
    mxArray *s = j2s_value(&r, emsg, sizeof(emsg));
    if (!s) { ssSetErrorStatus(S, strdup(emsg)); return; }

    /* export to ports */
    for(siglist = genomix->siglist, i = 0;
        siglist->type != GENOMIX_END;
        siglist++, i++) {
      double *port = ssGetOutputPortRealSignal(S, i);
      size = siglist->size;
      if (siglist->type == GENOMIX_SIZE)
        *port = 0.;

      if (s2p_value(s, siglist->type, siglist->name, genomix->dict,
                    &port, &size, emsg, sizeof(emsg)))
        ssSetErrorStatus(S, strdup(emsg));

      if (siglist->type == GENOMIX_VALUE)
        for(; size > 0; size--) *port++ = nan("");
    }

    mxDestroyArray(s);
  }
}

static void
mdlTerminate(SimStruct *S)
{
  struct genomix_port *genomix = ssGetPWorkValue(S, 0);
  struct genomix_dict *d;
  int i;

  if (genomix) {
    if (genomix->id >=0)
      genomix_disconnect(genomix->id);
    if (genomix->meta)
      mxDestroyArray(genomix->meta);
    if (genomix->json)
      free(genomix->json);
    if (genomix->siglist) {
      for(i = 0; genomix->siglist[i].type != GENOMIX_END; i++)
        free(genomix->siglist[i].name);
      free(genomix->siglist);
    }
    for(d = genomix->dict; d; d = genomix->dict) {
      genomix->dict = genomix->dict->next;
      mxFree(d->key);
      free(d);
    }
    free(genomix);
  }
}

void
genomix_event(const char *data)
{
}


/* --- getsiglist ---------------------------------------------------------- */

static struct genomix_siglist *
getsiglist(SimStruct *S)
{
  struct genomix_siglist *siglist;
  const mxArray *sigparam;
  const mxArray *p, *v;
  char key[256], emsg[256];
  const char *str;
  char *s;
  size_t i, nports;

  p = ssGetSFcnParam(S, GENOMIX_PARAM_SIGLIST);
  if (!p) { ssSetErrorStatus(S, "invalid arguments"); return NULL; }
  s = mxArrayToString(p);
  if (!s) { ssSetErrorStatus(S, "invalid arguments"); return NULL; }
  str = s;
  sigparam = j2s_value(&str, emsg, sizeof(emsg));
  mxFree(s);
  if (!sigparam) { ssSetErrorStatus(S, strdup(emsg)); return NULL; }

  nports = mxGetNumberOfElements(sigparam);
  siglist = malloc((1 + nports) * sizeof(*siglist));
  if (!siglist) { ssSetErrorStatus(S, "not enough memory"); return NULL; }

  for(i = 0; i < nports; i++) {
    p = mxGetCell(sigparam, i);
    if (!p) { ssSetErrorStatus(S, "invalid siglist"); return NULL; }

    v = mxGetField(p, 0, "type");
    if (!v) { ssSetErrorStatus(S, "no type key"); return NULL; }
    mxGetString(v, key, sizeof(key));
    if (!strcmp(key, "value"))
        siglist[i].type = GENOMIX_VALUE;
    else if (!strcmp(key, "size"))
        siglist[i].type = GENOMIX_SIZE;
    else {
      ssSetErrorStatus(S, "invalid type key"); return NULL;
    }

    v = mxGetField(p, 0, "name");
    if (!v) { ssSetErrorStatus(S, "no name key"); return NULL; }
    mxGetString(v, key, sizeof(key));
    siglist[i].name = malloc(strlen(key)+2);
    for(str = key, s = siglist[i].name; *str; str++, s++)
      *s = (*str == '.') ? 0 : *str;
    *s++ = 0;
    *s = 0;

    v = mxGetField(p, 0, "size");
    if (!v) { ssSetErrorStatus(S, "no size key"); return NULL; }
    siglist[i].size = mxGetScalar(v);
  }

  siglist[i].type = GENOMIX_END;
  siglist[i].name = NULL;
  siglist[i].size = 0;
  return siglist;
}


/* --- s2p_value ----------------------------------------------------------- */

/** Export required data into an output port.
 * Input is a matlab struct "data", which is recursively analyzed to find the
 * "field" member. Each element of "field" is converted to a number in the
 * "out" array (string values converted using the "dict" dictionary).
 */
static int
s2p_value(const mxArray *data, int type, const char *field,
          struct genomix_dict *dict,
          double **out, int *remain, char *emsg, size_t emsgsize)
{
  size_t i, m;

  if (!data) {
    snprintf(emsg, emsgsize, "invalid data for `%s'", field);
    return -1;
  }

  m = mxGetNumberOfElements(data);

  switch(mxGetClassID(data)) {
    case mxCELL_CLASS:
      for(i = 0; i < m; i++) {
        const mxArray *c = mxGetCell(data, i);

        if (c) {
          if (s2p_value(c, type, field, dict, out, remain, emsg, emsgsize))
            return -1;
        } else {
          if (type == GENOMIX_VALUE) {
            if (!*remain) return -1;
            *(*out)++ = nan("");
            (*remain)--;
          }
        }
      }
      break;

    case mxSTRUCT_CLASS:
      for(i = 0; i < m; i++) {
        if (*field) {
          const mxArray *f = mxGetField(data, i, field);
          if (!f) {
            snprintf(emsg, emsgsize, "cannot get field: `%s'", field);
            return -1;
          }
          if (s2p_value(f, type, field + strlen(field) + 1, dict, out, remain,
                        emsg, emsgsize))
            return -1;
        } else {
          int n = mxGetNumberOfFields(data);

          if (n == 0) {
            if (type == GENOMIX_VALUE) {
              if (!*remain) return -1;
              *(*out)++ = nan("");
              (*remain)--;
            }
          } else {
            int j;
            for(j = 0; j < n; j++) {
              if (s2p_value(mxGetFieldByNumber(data, i, j), type,
                            field, dict, out, remain, emsg, emsgsize))
                return -1;
            }
          }
        }
      }
      break;

    case mxLOGICAL_CLASS:
      if (*field) {
        snprintf(emsg, emsgsize, "no such field: `%s'", field);
        return -1;
      } else if (type == GENOMIX_SIZE) {
        *out += m;
      } else {
        mxLogical *v = mxGetData(data);

        for(i = 0; i < m; i++) {
          if (!*remain) return -1;
          *(*out)++ = *v++ ? 1.:0.;
          (*remain)--;
        }
      }
      break;

    case mxDOUBLE_CLASS:
      if (*field) {
        snprintf(emsg, emsgsize, "no such field: `%s'", field);
        return -1;
      } else if (type == GENOMIX_SIZE) {
        *out += m;
      } else {
        double *v = mxGetData(data);

        for(i = 0; i < m; i++) {
          if (!*remain) return -1;
          *(*out)++ = *v++;
          (*remain)--;
        }
      }
      break;

    case mxCHAR_CLASS:
      if (*field) {
        snprintf(emsg, emsgsize, "no such field: `%s'", field);
        return -1;
      } else if (type == GENOMIX_SIZE) {
        *out += 1;
      } else {
        char *t = mxArrayToString(data);

        **out = -1 /* invalid value by default */;
        if (t) {
          for(; dict; dict = dict->next)
            if (!strcmp(dict->key, t)) {
              **out = dict->value;
              break;
            }
        }
        (*out)++;
        (*remain)--;
      }
      break;

    default: break;
  }

  return 0;
}


/* --- p2s_value ----------------------------------------------------------- */

/** Import required data from an input port.
 * Input is a matlab struct "data", which is recursively analyzed to find the
 * "field" member. Each element of "field" is updated with a number from the
 * "in" array (or a string values converted using the "dict" dictionary).
 */
static int
p2s_value(mxArray *data, int type, const char *field,
          struct genomix_dict *dict,
          const double **in, int *remain, char *emsg, size_t emsgsize)
{
  size_t i, m;

  if (!data) {
    snprintf(emsg, emsgsize, "invalid data for `%s'", field);
    return -1;
  }

  if (!*field && type == GENOMIX_SIZE) {
    /* not supported atm */
    snprintf(emsg, emsgsize, "setting size not supported atm");
    return -1;
  }
  m = mxGetNumberOfElements(data);

  switch(mxGetClassID(data)) {
    case mxCELL_CLASS:
      for(i = 0; i < m; i++)
        if (p2s_value(mxGetCell(data, i), type, field, dict, in, remain,
                      emsg, emsgsize))
          return -1;
      break;

    case mxSTRUCT_CLASS:
      for(i = 0; i < m; i++) {
        if (*field) {
          mxArray *f = mxGetField(data, i, field);
          if (!f)
            return -1;
          if (p2s_value(f, type, field + strlen(field) + 1, dict, in, remain,
                        emsg, emsgsize))
            return -1;
        } else {
          int n = mxGetNumberOfFields(data);
          int j;

          for(j = 0; j < n; j++) {
            if (p2s_value(mxGetFieldByNumber(data, i, j), type,
                          field, dict, in, remain, emsg, emsgsize))
              return -1;
          }
        }
      }
      break;

    case mxDOUBLE_CLASS:
      if (*field) {
        snprintf(emsg, emsgsize, "no such field: `%s'", field);
        return -1;
      } else {
        double *v = mxGetPr(data);

        for(i = 0; i < m; i++) {
          if (!*remain) return -1;
          *v++ = *(*in)++;
          (*remain)--;
        }
      }
      break;

    case mxCHAR_CLASS:
      if (*field) {
        snprintf(emsg, emsgsize, "no such field: `%s'", field);
        return -1;
      } else {
        for(; dict; dict = dict->next)
          if (fabs(**in - dict->value) < 0.5) {
            mxArray *v = mxCreateString(dict->key);
            if (v) {
              /* swap data with new mxarray string */
              void *p = mxGetData(data);
              mxSetData(data, mxGetData(v));
              mxSetData(v, p);
              mxSetN(data, mxGetN(v));
              mxDestroyArray(v);
            }
            break;
          }
        if (!dict) {
          snprintf(emsg, emsgsize,
                   "invalid enum value %g%cfor: `%s'", **in,
                   *field ? ' ' : '\0', field);
          return -1;
        }
        (*in)++;
        (*remain)--;
      }
      break;

    default: break;
  }

  return 0;
}


/* --- m2s_input ----------------------------------------------------------- */

/** Convert meta data to matlab struct.
 * The returned mxArray is suitable to receive the block inputs so that it can
 * then directly be converted to json
 * Enums symbolic values are placed in the "dict" dictionary.
 */
static mxArray *
m2s_input(const mxArray *input, struct genomix_dict **dict)
{
  mxArray *value, *svalue, *stype;
  const char *field;
  int i, j, n;

  value = mxCreateStructArray(1, &(mwSize){1}, 0, NULL);
  if (!value) return NULL;

  n = mxGetNumberOfFields(input);
  for(i = 0; i < n; i++) {
    field = mxGetFieldNameByNumber(input, i);
    stype = mxGetFieldByNumber(input, 0, i);

    svalue = m2s_value(stype, dict);
    if (!svalue) return NULL;

    j = mxAddField(value, field);
    mxSetFieldByNumber(value, 0, j, svalue);
  }

  return value;
}

static mxArray *
m2s_value(const mxArray *format, struct genomix_dict **dict)
{
  char kind[64];
  const char *field;
  mxArray *value, *type, *svalue, *stype;
  mxArray *v;
  int i, j, n;
  double l;

  v = mxGetField(format, 0, "kind");
  if (!v) return NULL;
  mxGetString(v, kind, sizeof(kind));

  if (!strcmp(kind, "struct") || !strcmp(kind, "union")) {
    value = mxCreateStructArray(1, &(mwSize){1}, 0, NULL);
    if (!value) return NULL;

    type = mxGetField(format, 0, "type");
    if (!type) return NULL;

    n = mxGetNumberOfFields(type);
    for(i = 0; i < n; i++) {
      field = mxGetFieldNameByNumber(type, i);
      stype = mxGetFieldByNumber(type, 0, i);

      svalue = m2s_value(stype, dict);
      if (!svalue) return NULL;

      j = mxAddField(value, field);
      mxSetFieldByNumber(value, 0, j, svalue);
    }

  } else if (!strcmp(kind, "array") || !strcmp(kind, "sequence")) {
    v = mxGetField(format, 0, "length");
    if (v) l = mxGetScalar(v); else l = 16;

    value = mxCreateCellArray(2, (mwSize[2]){1, l});
    if (!value) return NULL;

    type = mxGetField(format, 0, "type");
    if (!type) return NULL;

    for(i = 0; i < l; i++) {
      svalue = m2s_value(type, dict);
      if (!svalue) return NULL;

      mxSetCell(value, i, svalue);
    }

  } else if (!strcmp(kind, "optional")) {
    type = mxGetField(format, 0, "type");
    if (!type) return NULL;

    value = m2s_value(type, dict);

  } else if (!strcmp(kind, "char") || !strcmp(kind, "string")) {
    value = mxCreateString("");

  } else if (!strcmp(kind, "enum")) {
    svalue = mxGetField(format, 0, "values");
    if (!svalue) return NULL;

    n = mxGetNumberOfElements(svalue);
    for(i = 0; i < n; i++) {
      char *t = NULL;
      v = mxGetCell(svalue, i);
      if (v) t = mxArrayToString(v);
      if (t) {
        struct genomix_dict *d = malloc(sizeof(*d));
        if (d) {
          mexMakeMemoryPersistent(t);
          *d = (struct genomix_dict){.key = t, .value = i, .next = *dict};
          *dict = d;
        }
      }
    }

    value = mxDuplicateArray(mxGetCell(svalue, 0));

  } else {
    value = mxCreateDoubleScalar(0.);
  }

  return value;
}

/* MEX-file interface mechanism */
#include "simulink.c"
